@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | PHP HTML Embedded Scripting Language Version 3.0                     |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997,1998 PHP Development Team (See Credits file)      |
+   | Copyright (c) 1997-2000 PHP Development Team (See Credits file)      |
    +----------------------------------------------------------------------+
    | This program is free software; you can redistribute it and/or modify |
    | it under the terms of one of the following licenses:                 |
@@ -26,17 +26,18 @@
    | Authors: Jim Winstead (jimw@php.net)                                 |
    +----------------------------------------------------------------------+
  */
+
+/* $Id: dbase.c,v 1.77 2000/01/01 04:31:14 sas Exp $ */
 #if defined(COMPILE_DL)
 #include "dl/phpdl.h"
-#undef DBASE
-#define DBASE 1
 #endif
 
-#if defined(THREAD_SAFE)
-#include "tls.h"
-#endif
 #include <stdlib.h>
 
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+
 #include "php.h"
 #include "internal_functions.h"
 #include "php3_list.h"
@@ -44,27 +45,11 @@
 #include "fopen-wrappers.h"
 
 #if DBASE
-#include "dbase/dbf.h"
-#if defined(THREAD_SAFE)
-DWORD DbaseTls;
-static int numthreads=0;
-void *dbase_mutex;
-
-typedef struct dbase_global_struct{
-	int le_dbhead;
-}dbase_global_struct;
-
-#define DBase_GLOBAL(a) dbase_globals->a
-
-#define DBase_TLS_VARS \
-	dbase_global_struct *dbase_globals; \
-	dbase_globals=TlsGetValue(DbaseTls); 
-
-#else
+#include "dbase.h"
+#include "../dbase/dbf.h"
 static int le_dbhead;
 #define DBase_GLOBAL(a) a
 #define DBase_TLS_VARS
-#endif
 
 #include <fcntl.h>
 #include <errno.h>
@@ -79,46 +64,16 @@ static void _close_dbase(dbhead_t *dbhead)
 
 int php3_minit_dbase(INIT_FUNC_ARGS)
 {
-#if defined(THREAD_SAFE)
-	dbase_global_struct *dbase_globals;
-#if !defined(COMPILE_DL)
-	CREATE_MUTEX(dbase_mutex,"DBase_TLS");
-	SET_MUTEX(dbase_mutex);
-	numthreads++;
-	if (numthreads==1){
-	if ((DbaseTls=TlsAlloc())==0xFFFFFFFF){
-		FREE_MUTEX(dbase_mutex);
-		return 0;
-	}}
-	FREE_MUTEX(dbase_mutex);
-#endif
-	dbase_globals = (dbase_global_struct *) LocalAlloc(LPTR, sizeof(dbase_global_struct)); 
-	TlsSetValue(DbaseTls, (void *) dbase_globals);
-#endif
 	DBase_GLOBAL(le_dbhead) = register_list_destructors(_close_dbase,NULL);
 	return SUCCESS;
 }
 
 static int php3_mend_dbase(void){
-#if defined(THREAD_SAFE)
-	dbase_global_struct *dbase_globals;
-	dbase_globals = TlsGetValue(DbaseTls); 
-	if (dbase_globals != 0) 
-		LocalFree((HLOCAL) dbase_globals); 
-#if !defined(COMPILE_DL)
-	SET_MUTEX(dbase_mutex);
-	numthreads--;
-	if (!numthreads){
-	if (!TlsFree(DbaseTls)){
-		FREE_MUTEX(dbase_mutex);
-		return 0;
-	}}
-	FREE_MUTEX(dbase_mutex);
-#endif
-#endif
 	return SUCCESS;
 }
 
+/* {{{ proto int dbase_open(string name, int mode)
+   Opens a dBase-format database file */
 void php3_dbase_open(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbf_name, *options;
 	dbhead_t *dbh;
@@ -148,7 +103,10 @@ void php3_dbase_open(INTERNAL_FUNCTION_PARAMETERS) {
 	handle = php3_list_insert(dbh, DBase_GLOBAL(le_dbhead));
 	RETURN_LONG(handle);
 }
+/* }}} */
 
+/* {{{ proto bool dbase_close(int identifier)
+   Closes an open dBase-format database file */
 void php3_dbase_close(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id;
 	dbhead_t *dbh;
@@ -168,7 +126,10 @@ void php3_dbase_close(INTERNAL_FUNCTION_PARAMETERS) {
 	php3_list_delete(dbh_id->value.lval);
 	RETURN_TRUE;
 }
+/* }}} */
 
+/* {{{ proto int dbase_numrecords(int identifier)
+   Returns the number of records in the database */
 void php3_dbase_numrecords(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id;
 	dbhead_t *dbh;
@@ -187,7 +148,10 @@ void php3_dbase_numrecords(INTERNAL_FUNCTION_PARAMETERS) {
 
 	RETURN_LONG(dbh->db_records);
 }
+/* }}} */
 
+/* {{{ proto int dbase_numfields(int identifier)
+   Returns the number of fields (columns) in the database */
 void php3_dbase_numfields(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id;
 	dbhead_t *dbh;
@@ -206,7 +170,10 @@ void php3_dbase_numfields(INTERNAL_FUNCTION_PARAMETERS) {
 
 	RETURN_LONG(dbh->db_nfields);
 }
+/* }}} */
 
+/* {{{ proto bool dbase_pack(int identifier)
+   Packs the database (deletes records marked for deletion) */
 void php3_dbase_pack(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id;
 	dbhead_t *dbh;
@@ -227,7 +194,10 @@ void php3_dbase_pack(INTERNAL_FUNCTION_PARAMETERS) {
         put_dbf_info(dbh);
 	RETURN_TRUE;
 }
+/* }}} */
 
+/* {{{ proto bool dbase_add_record(int identifier, array data)
+   Adds a record to the database */
 void php3_dbase_add_record(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id, *fields, *field;
 	dbhead_t *dbh;
@@ -292,7 +262,78 @@ void php3_dbase_add_record(INTERNAL_FUNCTION_PARAMETERS) {
 
 	RETURN_TRUE;
 }
+/* }}} */
 
+/* {{{ proto bool dbase_replace_record(int identifier, array data, int recnum)
+   Replaces a record to the database */
+PHP_FUNCTION(dbase_replace_record) {
+	pval *dbh_id, *fields, *field, *recnum;
+	dbhead_t *dbh;
+	int dbh_type;
+
+	int num_fields;
+	dbfield_t *dbf, *cur_f;
+	char *cp, *t_cp;
+	int i;
+	DBase_TLS_VARS;
+
+	if (ARG_COUNT(ht) != 3 || getParameters(ht,3,&dbh_id,&fields,&recnum)==FAILURE) {
+		WRONG_PARAM_COUNT;
+	}
+	convert_to_long(dbh_id);
+	convert_to_long(recnum);
+	if (fields->type != IS_ARRAY) {
+		php3_error(E_WARNING, "Expected array as second parameter");
+		RETURN_FALSE;
+	}
+
+	dbh = php3_list_find(dbh_id->value.lval, &dbh_type);
+	if (!dbh || dbh_type != DBase_GLOBAL(le_dbhead)) {
+		php3_error(E_WARNING, "Unable to find database for identifier %d", dbh_id->value.lval);
+		RETURN_FALSE;
+	}
+
+	num_fields = _php3_hash_num_elements(fields->value.ht);
+
+	if (num_fields != dbh->db_nfields) {
+		php3_error(E_WARNING, "Wrong number of fields specified");
+		RETURN_FALSE;
+	}
+
+	cp = t_cp = (char *)emalloc(dbh->db_rlen + 1);
+	if (!cp) {
+		php3_error(E_WARNING, "unable to allocate memory");
+		RETURN_FALSE;
+	}
+	*t_cp++ = VALID_RECORD;
+
+	dbf = dbh->db_fields;
+	for (i = 0, cur_f = dbf; cur_f < &dbf[num_fields]; i++, cur_f++) {
+		if (_php3_hash_index_find(fields->value.ht, i, (void **)&field) == FAILURE) {
+			php3_error(E_WARNING, "unexpected error");
+			efree(cp);
+			RETURN_FALSE;
+		}
+		convert_to_string(field);
+		sprintf(t_cp, cur_f->db_format, field->value.str.val); 
+		t_cp += cur_f->db_flen;
+	}
+
+	if (put_dbf_record(dbh, recnum->value.lval, cp) < 0) {
+		php3_error(E_WARNING, "unable to put record at %ld", dbh->db_records);
+		efree(cp);
+		RETURN_FALSE;
+	}
+
+        put_dbf_info(dbh);
+	efree(cp);
+
+	RETURN_TRUE;
+}
+/* }}} */
+
+/* {{{ proto bool dbase_delete_record(int identifier, int record)
+   Marks a record to be deleted */
 void php3_dbase_delete_record(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id, *record;
 	dbhead_t *dbh;
@@ -323,13 +364,17 @@ void php3_dbase_delete_record(INTERNAL_FUNCTION_PARAMETERS) {
         put_dbf_info(dbh);
 	RETURN_TRUE;
 }
+/* }}} */
 
+/* {{{ proto array dbase_get_record(int identifier, int record)
+   Returns an array representing a record from the database */
 void php3_dbase_get_record(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id, *record;
 	dbhead_t *dbh;
 	int dbh_type;
 	dbfield_t *dbf, *cur_f;
 	char *data, *fnp, *str_value;
+	size_t cursize = 0;
 	DBase_TLS_VARS;
 
 	if (ARG_COUNT(ht) != 2 || getParameters(ht,2,&dbh_id,&record)==FAILURE) {
@@ -355,11 +400,17 @@ void php3_dbase_get_record(INTERNAL_FUNCTION_PARAMETERS) {
 		RETURN_FALSE;
 	}
 
-        fnp = (char *)emalloc(dbh->db_rlen);
+		fnp = NULL;
         for (cur_f = dbf; cur_f < &dbf[dbh->db_nfields]; cur_f++) {
 		/* get the value */
 		str_value = (char *)emalloc(cur_f->db_flen + 1);
-                sprintf(str_value, cur_f->db_format, get_field_val(data, cur_f, fnp));
+
+		if(cursize <= cur_f->db_flen) {
+			cursize = cur_f->db_flen + 1;
+			fnp = erealloc(fnp, cursize);
+		}
+		
+                snprintf(str_value, cursize, cur_f->db_format, get_field_val(data, cur_f, fnp));
 
 		/* now convert it to the right php internal type */
 	        switch (cur_f->db_type) {
@@ -397,8 +448,11 @@ void php3_dbase_get_record(INTERNAL_FUNCTION_PARAMETERS) {
 
 	free(data);
 }
+/* }}} */
 
 /* From Martin Kuba <makub@aida.inet.cz> */
+/* {{{ proto array dbase_get_record_with_names(int identifier, int record)
+   Returns an associative array representing a record from the database */
 void php3_dbase_get_record_with_names(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *dbh_id, *record;
 	dbhead_t *dbh;
@@ -470,7 +524,10 @@ void php3_dbase_get_record_with_names(INTERNAL_FUNCTION_PARAMETERS) {
 
 	free(data);
 }
+/* }}} */
 
+/* {{{ proto bool dbase_create(string filename, array fields)
+   Creates a new dBase-format database file */
 void php3_dbase_create(INTERNAL_FUNCTION_PARAMETERS) {
 	pval *filename, *fields, *field, *value;
 	int fd;
@@ -499,7 +556,7 @@ void php3_dbase_create(INTERNAL_FUNCTION_PARAMETERS) {
 		RETURN_FALSE;
 	}
 
-	if ((fd = open(filename->value.str.val, O_RDWR|O_CREAT, 0644)) < 0) {
+	if ((fd = open(filename->value.str.val, O_BINARY|O_RDWR|O_CREAT, 0644)) < 0) {
 		php3_error(E_WARNING, "Unable to create database (%d): %s", errno, strerror(errno));
 		RETURN_FALSE;
 	}
@@ -613,6 +670,7 @@ void php3_dbase_create(INTERNAL_FUNCTION_PARAMETERS) {
 	handle = php3_list_insert(dbh, DBase_GLOBAL(le_dbhead));
 	RETURN_LONG(handle);
 }
+/* }}} */
 
 function_entry dbase_functions[] = {
 	{"dbase_open",			php3_dbase_open,			NULL},
@@ -622,9 +680,10 @@ function_entry dbase_functions[] = {
 	{"dbase_numfields",		php3_dbase_numfields,		NULL},
 	{"dbase_add_record",	php3_dbase_add_record,		NULL},
 	{"dbase_get_record",	php3_dbase_get_record,		NULL},
-       {"dbase_get_record_with_names", php3_dbase_get_record_with_names,               NULL},
+	{"dbase_get_record_with_names", php3_dbase_get_record_with_names,               NULL},
 	{"dbase_delete_record",	php3_dbase_delete_record,	NULL},
 	{"dbase_pack",			php3_dbase_pack,			NULL},
+	PHP_FE(dbase_replace_record, NULL)
 	{NULL, NULL, NULL}
 };
 
@@ -635,18 +694,6 @@ php3_module_entry dbase_module_entry = {
 
 #if defined(COMPILE_DL)
 DLEXPORT php3_module_entry *get_module(void) { return &dbase_module_entry; }
-
-#if (WIN32|WINNT) && defined(THREAD_SAFE)
-
-/*NOTE: You should have an odbc.def file where you
-export DllMain*/
-BOOL WINAPI DllMain(HANDLE hModule, 
-                      DWORD  ul_reason_for_call, 
-                      LPVOID lpReserved)
-{
-    return 1;
-}
-#endif
 #endif
 
 #endif
